from setuptools import setup

"""
File name: setup
Date created: 19/02/2019
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

setup(name='gym_elecsim',
      version='0.0.1',
      install_requires=['gym']  # And any other dependencies foo needs
)
